#!/usr/bin/env python

from Common import *
from UserDict import UserDict
from Configurable import Configurable
from ExceptionHandler import ExceptionHandler
from Servlet import Servlet
from ServletFactory import *
from types import FloatType

try:
	import WebUtils
except:
	sys.path.append('..')
	import WebUtils
from WebUtils.WebFuncs import HTMLEncode
from WebUtils.HTMLForException import HTMLForException

# @@ import MiddleKit
# @@ from MiddleKit.KeyValueAccess import KeyValueAccess
import MiscUtils.CSV

# Beef up UserDict with the KeyValueAccess base class and custom versions of
# hasValueForKey() and valueForKey(). This all means that UserDict's (such as
# os.environ) are key/value accessible. At some point, this probably needs to
# move somewhere else as other Webware components will need this "patch".
# (@@ 2000-05-07 ce: Plus I just duplicated this from CGIWrapper.py. Doh!)
import MiddleKit
from MiddleKit.KeyValueAccess import KeyValueAccess
if not KeyValueAccess in UserDict.__bases__:
	UserDict.__bases__ = UserDict.__bases__ + (KeyValueAccess,)

	def _UserDict_hasValueForKey(self, key):
		return self.has_key(key)

	def _UserDict_valueForKey(self, key, default=None): # @@ 2000-05-10 ce: does Tombstone fit here and possibly in KeyValueAccess?
		return self.get(key, default)

	setattr(UserDict, 'hasValueForKey', _UserDict_hasValueForKey)
	setattr(UserDict, 'valueForKey', _UserDict_valueForKey)

	
class ApplicationError(Exception):
	pass
	
		
class Application(Configurable):
	'''
	FUTURE
		* 2000-04-09 ce: Actions: URLs that are bound to specific methods of a servlet
			  - These could be powerful, because they could send the request back to the
				same object that generated the page, but invoke a different method.
				That's useful for at least forms processing and it's kind of like buttons
				on a GUI window that tie back to the same controller.
		* 2000-04-09 ce: Automatically open in browser.
		* 2000-04-09 ce: Option to remove HTML comments in responses.
		* 2000-04-09 ce: Option remove unnecessary white space in responses.
		* 2000-04-09 ce: Debugging flag and debug print method.
		* 2000-04-09 ce: A web-based, interactive monitor to the application.
		* 2000-04-09 ce: Record and playback of requests and responses. Useful for regression testing.
		* 2000-04-09 ce: sessionTimeout() and a hook for when the session has timed out.
		* 2000-04-09 ce: pageRefreshOnBacktrack
		* 2000-04-09 ce: terminate() and isTerminating()
		* 2000-04-09 ce: isRefusingNewSessions()
		* 2000-04-09 ce: terminateAfterTimeInterval()
		* 2000-04-09 ce: restoreSessionWithID:inTransaction:
		* 2000-04-09 ce: pageWithNameForRequest/Transaction() (?)
		* 2000-04-09 ce: port() and setPort() (?)
		* 2000-04-09 ce: Adaptors (?)
		* 2000-04-09 ce: Does request handling need to be embodied in a separate object?
			  - Probably, as we may want request handlers for various file types.
		* 2000-04-09 ce: Concurrent request handling (probably through multi-threading)
	'''
	
	## Init ##
	
	def __init__(self, server=None, transactionClass=None, sessionClass=None, requestClass=None, responseClass=None, exceptionHandlerClass=None):
		Configurable.__init__(self)

		self._server = server

		if transactionClass:
			self._transactionClass = transactionClass
		else:
			from Transaction import Transaction
			self._transactionClass = Transaction
		
		if sessionClass:
			self._sessionClass = sessionClass
		else:
			from Session import Session
			self._sessionClass = Session
		
		if requestClass:
			self._requestClass = requestClass
		else:
			from HTTPRequest import HTTPRequest
			self._requestClass = HTTPRequest

		if responseClass:
			self._responseClass = responseClass
		else:
			from HTTPResponse import HTTPResponse
			self._responseClass = HTTPResponse

		if exceptionHandlerClass:
			self._exceptionHandlerClass = exceptionHandlerClass
		else:
			from ExceptionHandler import ExceptionHandler
			self._exceptionHandlerClass = ExceptionHandler

		self._sessions = {}

		self._servletCacheByPath = {}

		self._serverDir = os.getcwd()
		
		self._factoryList = [PythonServletFactory(self), PSPServletFactory(self), UnknownFileTypeServletFactory(self)]
		# ^ @@ 2000-05-03 ce: make this customizable at least through a method (that can be overridden) if not a config file (or both)
	
		# Create a dictionary that maps all known extensions to their factories, for quick look up
		self._factoryByExt = {}
		for factory in self._factoryList:
			assert isinstance(factory, ServletFactory)
			for ext in factory.extensions():
				assert not self._factoryByExt.has_key(ext), 'Extension (%s) for factory (%s) was already used by factory (%s)' % (ext, self._factoryByExt[ext].name(), factory.name())
				self._factoryByExt[ext] = factory
		
		if self.setting('PrintConfigAtStartUp'):
			self.printConfig()

		print 'Current directory is', os.getcwd()


	## Config ##
	
	def defaultConfig(self):
		return {
			'PrintConfigAtStartUp': 1,
			'ServletsDir':          'Examples',
			'LogActivity':          1,
			'ActivityLogFilename':  'Logs/Activity.csv',
			'ActivityLogColumns':   ['request.remoteAddress', 'request.method', 'request.uri', 'response.size', 'servlet.name', 'request.timeStamp', 'transaction.duration', 'transaction.errorOccurred'],
			
			# Error handling
			'ShowDebugInfoOnErrors':  1,
			'UserErrorMessage':       'The site is having technical difficulties with this page. An error has been logged, and the problem will be fixed as soon as possible. Sorry!',
			'ErrorLogFilename':       'Logs/Errors.csv',
			'SaveErrorMessages':      1,
			'ErrorMessagesDir':       'ErrorMsgs',
			'EmailErrors':            0, # be sure to review the following settings when enabling error e-mails
			'ErrorEmailServer':       'mail.-.com',
			'ErrorEmailHeaders':      { 'From':         '-@-.com',
				                        'To':           ['-@-.com'],
				                        'Reply-to':     '-@-.com',
				                        'Content-type': 'text/html',
				                        'Subject':      'Error'
									}
		}
	
	def configFilename(self):
		return 'Configs/Application.config'


	## Versions ##
	
	def webKitVersion(self):
		return '0.2'
		
	def version(self):
		''' Returns the version of the application. This implementation returns '0.1'. Subclasses should override to return the correct version number. '''
		## @@ 2000-05-01 ce: Maybe this could be a setting 'AppVersion'
		return '0.1'

			
	## Dispatching Requests ##
	
	def dispatchRawRequest(self, newRequestDict):
		return self.dispatchRequest(self.createRequestForDict(newRequestDict))
	
	def dispatchRequest(self, newRequest):
		''' Creates the transaction, session, response and servlet for the new request which is then dispatched. The transaction is returned. '''
		transaction = None
		try:
			request     = newRequest
			transaction = self.createTransactionForRequest(request)
			session     = self.createSessionForTransaction(transaction)
			response    = self.createResponseInTransaction(transaction)
			self.createServletInTransaction(transaction)

			self.awake(transaction)
			self.respond(transaction)
			self.sleep(transaction)
			
			transaction.response().deliver(transaction)
		
		except:
			if transaction:
				transaction.setErrorOccurred(1)
			self.handleExceptionInTransaction(sys.exc_info(), transaction)
 			transaction.response().deliver(transaction) # I hope this doesn't throw an exception. :-)   @@ 2000-05-09 ce: provide a secondary exception handling mechanism

		if self.setting('LogActivity'):
			self.writeActivityLog(transaction)
		
		return transaction


	## Transactions ##

	# @@ 2000-05-10 ce: should just send the message to the transaction and let it handle the rest.
	
	def awake(self, transaction):
		transaction.session().awake(transaction)
		transaction.servlet().awake(transaction)
	
	def respond(self, transaction):
		transaction.session().respond(transaction)
		transaction.servlet().respond(transaction)
	
	def sleep(self, transaction):
		transaction.session().sleep(transaction)
		transaction.servlet().sleep(transaction)


	## Sessions ##
	
	def session(self, sessionId, default=Tombstone):
		if default is Tombstone:
			return self._sessions[sessionId]
		else:
			return self._sessions.get(sessionId, default)

	def hasSession(self, sessionId):
		return self._sessions.has_key(sessionId)

	def sessions(self):
		return self._sessions

		
	## Misc Access ##

	def server(self):
		return self._server

	def name(self):
		return sys.argv[0]
	
	def transactionClass(self):
		return self._transactionClass
	
	def setTransactionClass(self, newClass):
		assert isclass(newClass)
		self._transactionClass = newClass
	
	def responseClass(self, newClass):
		return self._responseClass

	def setResponseClass(self, newClass):
		assert isclass(newClass)
		self._responseClass = newClass


	## Activity Log ##
	
	def writeActivityLog(self, transaction):
		''' Writes an entry to the script log file. Uses settings ActivityLogFilename and ActivityLogColumns. '''
		filename = os.path.join(self._serverDir, self.setting('ActivityLogFilename'))
		if os.path.exists(filename):
			file = open(filename, 'a')
		else:
			file = open(filename, 'w')
			file.write(string.join(self.setting('ActivityLogColumns'), ',')+'\n')
		values = []
		# We use UserDict on the next line because we know it inherits KeyValueAccess and reponds to valueForName()
		objects = UserDict({
			'application': self,
			'transaction': transaction,
			'request':     transaction.request(),
			'response':    transaction.response(),
			'servlet':     transaction.servlet(),
			'session':     transaction.session()
		})
		for column in self.setting('ActivityLogColumns'):
			value = objects.valueForName(column)
			if type(value) is FloatType:
				value = '%0.2f' % value   # probably need more flexibility in the future
			else:
				value = str(value)
			values.append(value)
		file.write(string.join(values, ',')+'\n')
		file.close()


	## Utilities/Hooks ##
	
	def serverDir(self):
		''' Returns the directory where the application server is located. '''
		return self._serverDir

	def createRequestForDict(self, newRequestDict):
		return self._requestClass(dict=newRequestDict)

	def createTransactionForRequest(self, request):
		trans = self._transactionClass(application=self, request=request)
		request.setTransaction(trans)
		return trans

	def createResponseInTransaction(self, transaction):
		response = self._responseClass()
		transaction.setResponse(response)
		return response

	def createSessionForTransaction(self, transaction):
		sessId = transaction.request().sessionId()
		session = None
		if sessId:
			session = self.session(sessId, None)
		# @@ 2000-05-10 ce: So here we just make a new session if the existing session id is invalid. Is that really the behavior we want? At the very least, we should probably have a "hook" method that subclasses can customize
		if session is None:
			session = self._sessionClass(transaction)
			self._sessions[session.identifier()] = session
		transaction.setSession(session)
		return session

	def createServletInTransaction(self, transaction):
		# Get the path, extension and factory
		path = transaction.request().serverSidePath()
		ext = os.path.splitext(path)[1]
		factory = self._factoryByExt.get(ext, None)
		if not factory:
			factory = self._factoryByExt.get('.*', None) # special case: .* is the catch-all
			if not factory:
				raise ApplicationError, 'Unknown extension (%s). No factory found.' % ext
				# ^ @@ 2000-05-03 ce: Maybe the web browser doesn't want an exception for bad extensions. We probably need a nicer message to the user...
				#                     On the other hand, that can always be done by providing a factory for '.*'
		assert factory.uniqueness()=='file', '%s uniqueness is not supported.' % factory.uniqueness()

		# Cached?
		cache = self._servletCacheByPath.get(path, None)
		
		# File is not newer?
		if cache and self._servletCacheByPath[path]['timestamp']<os.path.getmtime(path):
			cache = None
			
		# Instance can be reused?
		if cache and not cache['instance'].canBeReused():
			cache = None

		# Create the cache?		
		if not cache:
			# Add the path to sys.path. @@ 2000-05-09 ce: not the most ideal solution, but works for now
			dir = os.path.split(path)[0]
			
			if not dir in sys.path:
				sys.path.insert(0, dir)

			cache = {
				'instance':  factory.servletForTransaction(transaction),
				'path':      path,
				'timestamp': os.path.getmtime(path)
			}
			assert cache['instance'] is not None, 'Factory (%s) failed to create a servlet upon request.' % factory.name()
			self._servletCacheByPath[path] = cache

		# Set the transaction's servlet
		transaction.setServlet(cache['instance'])
	
	def handleExceptionInTransaction(self, excInfo, transaction):
		self._exceptionHandlerClass(self, transaction, excInfo)


def main(requestDict):
	from WebUtils.HTMLForException import HTMLForException
	try:
		assert type(requestDict) is type({})
		from HTTPRequest import HTTPRequest
		request = HTTPRequest(requestDict)
		if 0:
			# For debugging in bad circumstances
			print '<html><body>'
			print '<p><b>Application %s</b>' % sys.argv[1]
			print request.htmlInfo()
			print '</body></html>'
		app = Application()
		app.dispatchRequest(request)
	except:
		print 'Content-type: text/html\n'
		print '<html>'
		print '<body>'
		print HTMLForException()
		print '</body>'
		print '</html>'


# You can run Application as a main script, in which case it expects a single
# argument which is a file containing a dictionary representing a request. This
# technique isn't very popular as Application itself could raise exceptions
# that aren't caught. See CGIAdaptor.py and AppServer.py for a better example of
# how things should be done.
if __name__=='__main__':
	if len(sys.argv)!=2:
		sys.stderr.write('WebKit: Application: Expecting one filename argument.\n')
	requestDict = eval(open(sys.argv[1]).read())
	main(requestDict)
