from Common import *
from HTTPServlet import HTTPServlet


class Page(HTTPServlet):
	'''
	Page is a type of HTTPServlet that is more convenient for Servlets which represent HTML pages generated in response to GET and POST requests. In fact, this is the most common type of Servlet.
	Subclasses typically override writeHeader(), writeBody() and writeFooter().
	They might also choose to override writeHTML() entirely.
	In awake(), the page sets self attributes, _transaction, _response and _request which subclasses should use as appropriate.
	For the purposes of output, the write() and writeln() convenience methods are provided.
	When developing a full-blown website, it's common to create a subclass of Page called SitePage which defines the common look and feel of the website and provides site-specific convenience methods. Then all other page classes inherit from SitePage.
	'''

	super = HTTPServlet

	
	## Transactions ##
	
	def awake(self, transaction):
		self.super.awake(self, transaction)
		# @@ 2000-05-08 ce: move these up to Servlet?
		self._transaction  = transaction
		self._response = transaction.response()
		self._request  = transaction.request()
		self._session  = transaction.session()
		assert self._transaction  is not None
		assert self._response is not None
		assert self._request  is not None
		# @@ 2000-05-08 ce: yes/no: assert self._session is not None

	def respondToGet(self, transaction):
		''' Invokes writeHTML(). '''
		self.writeHTML()
	
	def respondToPost(self, transaction):
		''' Invokes writeHTML(). '''
		self.writeHTML()

	def sleep(self, transaction):
		self._transaction  = None
		self._response = None
		self._request  = None
		self.super.sleep(self, transaction)


	## Access ##
	
	def application(self):
		return self._transaction.application()

	def transaction(self):
		return self._transaction
		
	def request(self):
		return self._request
		
	def response(self):
		return self._response

				
	## Generating results ##
	
	def title(self):
		''' Subclasses often override this method to provide a custom title. This title should be absent of HTML tags. This implementation returns the name of the class, which is sometimes appropriate and at least informative. '''
		return self.__class__.__name__

	def htTitle(self):
		''' Return self.title(). Subclasses sometimes override this to provide an HTML enhanced version of the title. This is the method that should be used when including the page title in the actual page contents. '''
		return self.title()

	def htBodyArgs(self):
		''' The arguments used for the HTML <body> tag. Invoked by writeHeader(). '''
		return 'color=black bgcolor=white'

	def writeHTML(self):
		''' Subclasses may override this method (which is invoked by respondToGet() and respondToPost()) or it's constituent methods, writeHeader(), writeBody() and writeFooter(). '''
		self.writeln('<html>')
		self.writeHeader()
		self.writeBody()
		self.writeFooter()
		self.writeln('</html>')

	def writeHeader(self):
		self.writeln('''<head>
	<title>%s</title>
</head>
<body %s>''' % (self.title(), self.htBodyArgs()))

	def writeBody(self):
		self.writeln("<p>This page has not yet customized it's body.")
		
	def writeFooter(self):
		self.writeln('</body>')
	

	## Utility methods ##

	def write(self, *args):
		for arg in args:
			self._response.write(str(arg))

	def writeln(self, *args):
		for arg in args:
			self._response.write(str(arg))
		self._response.write('\n')

		
	## Threading ##
	
	def canBeThreaded(self):
		''' Returns 0 because of the ivars we set up in awake(). '''
		return 0
